//------------------------------------------------------------------------
//
// OpenDCC - S88-N - Keyboard and RC5 Receiver
//
// Copyright (c) 2007 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      main.c
// author:    Wolfgang Kufer (main)
//            RC5 code taken from Peter Dannegger, danni@specs.de 
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2007-11-27 V0.01 start
// history:   2007-11-27 V0.04 added 4x4 matrix (see s88_key.c)
//            2009-12-28 V0.05 bugfix in jumper detection
//            2010-02-02 V0.06 bugfix in current_message (atomic)
//
//------------------------------------------------------------------------
//
// purpose:   keyboard and RC5 feedback system
//
// content:   
//             1. Defines and variable definitions
//
// interrupts:  USI_OVERFLOW (for s88 handling)
//              TIMER_COMPA (for timetick and rc5 analysis)
//
// timing:    on s88:
//            from falling clk edge until update of port:
//               t-min: 10us;
//               t-max: 15us (with extra sei() in RC5 interrupt routine)
//               t-max: 25us (normal interrupt routine)
//               these values are not visible, since we shift out data
//               by hardware.
//
//------------------------------------------------------------------------
// Ideen fr einen GBM (gemailt an Blcher 10.01.08):
//
// 1. Ethernetschnittstelle, dafr dann auch ein Programm,
//    das sich als serieller Treiber im PC einnistet und zum
//    Anwendungsprogramm hin ein HSI88 vorgaukelt.
//    Damit wrde Ihre Schnittstelle nahtlos von allen
//    PC-Programmen untersttzt werden.
// 2. Virtueller Melder fr DCC-Signal  Boosterausfall
// 2a) Meldung, welches Gleis Kurzschlu verursacht hat.
// 3. Des weiteren ein virtueller Melder, der vor sich hin blinkert
//    - damit kann der PC sehen, ob das noch alles glaubhaft ist.
// 4. Summenstrommessung: - aber vielleicht ist da ein A/D am werkeln:
//    da knnte man den Summenstrom doch auch rausfinden und melden,
//    bzw. gegen Schwellwerte vergleichen und deren berschreitung melden.
// 5. Integrierte Kehrschleife (3 Melder umschaltbar bzw. Relaisausgang,
//    an jedem Ende die Schaltreferenzmelder whlbar.
// Zusammengefasst also folgender Vorschlag:
//
//  GBM belegt 32 Melderadressen
//  16 echte Sensormelder
//  16 virtuelle Melder:
//    - 1 DCC Ausfallmelder.
//    - 1 blinkender Melder (z.B. 0,5Hz)
//    - 8 Thresholdmelder Summenstrom
//    - Rest reserved

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <string.h>

#include "config.h"
#include "s88_key.h"
#include "main.h"


#define SIMULATION  0            // 0: real application
                                 // 1: ???

#define SHOW_S88_ON_LED        FALSE    // TRUE: turn on LED each s88 clk
#define SHOW_IR_BITS_ON_LED    FALSE    // TRUE: turn on LED after filtering
#define SHOW_IR_TOGGLE_ON_LED  FALSE    // TRUE: turn on LED each toggle

#define RC5_METHOD   2              // 0 = void
                                    // 1 = Dannegger Code - einfache Timinganalyse
                                    // 2 = Timinganalyse mit Filter und
                                    //     Halbbitauswertung
                                    // 3 = Multicode fr versch. Codecs - unvollstndig!

#define USE_FILTER_ON_RC5  TRUE     // TRUE: run a noise filter on input
                                    //       consumes some CPU time
                                    //       but hardens against noise
                                    // FALSE: Read input directly

#ifndef RC5_PORT
 #define	RC5_PORT        PIND
 #define	RC5_PIN         PD2	    // IR input low active
#endif


//
//-------------------------------------------------------------------------
//
// Connect the Attiny2313 as follows:
//
//
//                       --------U--------
//       (reset)     NC  | 1 PA2   VCC 20|-
//   +--|>|-- 2k -- LED  | 2 PD0   PB7 19|- S88_CLK
//            S88_RESET  | 3 PD1   PB6 18|- S88_DO
//       (xtal) JUMPER2  | 4 PA1   PB5 17|- S88_DI
//       (xtal) JUMPER1  | 5 PA0   PB4 16|- S88_LOAD
//         RC5_receiver  | 6 PD2   PB3 15|- KEY 8
//                 KEY1  | 7 PD3   PB2 14|- KEY 7
//                 KEY2  | 8 PD4   PB1 13|- KEY 6
//                 KEY3  | 9 PD5   PB0 12|- KEY 5
//                  GND  |10 GND   PD6 11|- KEY 4
//                       -----------------
//
//
//----------------------------------------------------------------------------

t_keyboard_type keyboard_type;

volatile signed char timerval;    // is incremented by ISR

volatile unsigned int current_message;


void init_main(void)
  {
    DDRA  =  0;                  // all in
    PORTA = (1 << JUMPER1)
          | (1 << JUMPER2);      // pullup
    
    DDRB  =  0                   // PortB: 1=out, 0=in
          | (0 << PS88_CLK)  
          | (1 << PS88_DO)
          | (0 << PS88_DI)
          | (0 << PS88_LOAD);
    PORTB  = 0b00001111          // PortB: 1=pull up
          | (1 << PS88_CLK)  
          | (0 << PS88_DO)
          | (1 << PS88_DI)
          | (1 << PS88_LOAD);

    DDRD  = (1 << LED);         // else: all input

    PORTD = 0b01111000          // pullup - set to 1
          | (1 << LED)          // turn off LED          
          | (1 << PS88_RESET) 
          | (1 << RC5_n);

    // Init Timer1

    TCCR1B = (0 << ICNC1) 
           | (0 << ICES1) 
           | (0 << WGM13) 
           | (1 << WGM12)       // Mode 4: CTC with OCRA1
           | (0 << CS12)        // clk stopped
           | (0 << CS11) 
           | (0 << CS10);


    // Init Timer0 as CTC 
    // check TICK_PERIOD and F_CPU
    #define TIMER0_CLKDIV 8              // possible values: 1, 8, 64, 256, 1024

    #if (F_CPU / 1000000L * TICK_PERIOD / TIMER0_CLKDIV) > 255L
      #warning: overflow in OCR0A - check TICK_PERIOD and F_CPU
      #warning: suggestion: use a larger clkdiv
    #endif    
    #if (F_CPU / 1000000L * TICK_PERIOD / TIMER0_CLKDIV) < 30L
      #warning: resolution accuracy in OCR0A too low - check TICK_PERIOD and F_CPU
      #warning: suggestion: use a smaller clkdiv
    #endif    


    OCR0A = F_CPU / 1000000L * TICK_PERIOD / TIMER0_CLKDIV ;  
    TCCR0A = (0 << COM0A1)      // compare match A
           | (0 << COM0A0) 
           | (0 << COM0B1)      // compare match B
           | (0 << COM0B0) 
           | 0                  // reserved
           | 0                  // reserved
           | (1 << WGM01)  
           | (0 << WGM00);      // Timer0 Mode 2 = CTC - Int on compare A
    TCCR0B = (0 << FOC0A) 
           | (0 << FOC0B) 
           | (0 << WGM02) 
    #if   TIMER0_CLKDIV == 1
           | (0 << CS02) | (0 << CS01)  | (1 << CS00);       // CS[2:0]=001 clkdiv 1
    #elif TIMER0_CLKDIV == 8
           | (0 << CS02) | (1 << CS01)  | (0 << CS00);       // CS[2:0]=010 clkdiv 8
    #elif TIMER0_CLKDIV == 64
           | (0 << CS02) | (1 << CS01)  | (1 << CS00);       // CS[2:0]=011 clkdiv 64 
    #elif TIMER0_CLKDIV == 256
           | (1 << CS02) | (0 << CS01)  | (0 << CS00);       // CS[2:0]=100 clkdiv 256
    #elif TIMER0_CLKDIV == 1024
           | (1 << CS02) | (0 << CS01)  | (1 << CS00);       // CS[2:0]=101 clkdiv 1024   
    #else
     #warning: TIMER0_CLKDIV is void
    #endif


    TIMSK = (0<<TOIE1)       // Timer1 Overflow
          | (0<<OCIE1A)      // Timer1 Compare A
          | (0<<OCIE1B)      // Timer1 Compare B
          | 0                // reserved
          | (0<<ICIE1)       // Timer1 Input Capture
          | (0<<OCIE0B)      // Timer0 Compare B
          | (0<<TOIE0)       // Timer0 Overflow
          | (1<<OCIE0A);     // Timer0 Compare A -> used here


    // Init Interrupt 0

    GIMSK = (0<<INT1)        // Int1 is msb
          | (0<<INT0)        // INT0 disabled
          | (0<<PCIE); 

    MCUCR = 0x00;  //           // no interrupts ;The rising edge of INT0 generates an interrupt request.


    // USI

    USICR = (0 << USISIE)    // Start Condition Interrupt Enable   
          | (1 << USIOIE)    // Counter Overflow Interrupt Enable       
          | (0 << USIWM1)    // USIWM1..0: Wire Mode     
          | (1 << USIWM0)    // 00: normal Mode
                             // 01: three wire mode (used here)
                             // 10: two wire mode
          | (1 << USICS1)    // Clock Source:    
          | (0 << USICS0)    // 10: external, pos. edge    
          | (0 << USICLK)    // Clkout (not used here)
          | (0 << USITC);    // Toggle        
         
    USISR = 0b00001111;      // set Counter to 15, to get an interrupt on next clk change   
  }


unsigned char copy[] PROGMEM = {".... S88_RC5 V0.06 (c) Kufer 2007-2010 ..."};


//---------------------------------------------------------------------------------------------

#if (RC5_ENABLED == TRUE)
// RC5 Variablen

#define RC5TIME     1778L		// 1.778msec

// An RC5 frame has 14 bits:
// S1 S2[~C.6] T A4 A3 A2 A1 A0 C5 C4 C3 C2 C1 C0
//      S: automatic Gain control bits (start bit)
//      T: toggle bit
//      A: device address
//      C: command, complement of S2 is seventh command bit

#define PULSE_MIN   (RC5TIME *  4L / 10L)
#define PULSE_1_2   (RC5TIME *  8L / 10L)
#define PULSE_MAX	(RC5TIME * 12L / 10L)

#define	xRC5_IN		PIND
#define	xRC5		PD2			// IR input low active


struct
  {
    enum
      {
        IDLE,
        ZERO_1ST_HALF,
        ZERO_2ND_HALF,
        ONE_1ST_HALF,
        ONE_2ND_HALF
      } state;
    unsigned char   received;           // Flag
    unsigned char   filter_data;        // sampler

    unsigned char	bit;				// bit value
    unsigned char	time;				// count bit time
    unsigned int	tmp;				// shift bits in
    unsigned int	data;				// store result
  } rc5;

#endif

static inline void rc5_sampler(void) 
       __attribute__((always_inline));


ISR(TIMER0_COMPA_vect)
  {
    timerval++;
    rc5_sampler();    // now handle RC5 Code
  }


#if (RC5_ENABLED == TRUE)

#if (RC5_METHOD == 1)                               // Code according to dannegger
    void rc5_sampler(void)
      {
        unsigned int tmp = rc5.tmp;				    // for faster access

        if ( ++rc5.time > (PULSE_MAX / TICK_PERIOD) )
          {			                                // count pulse time
            if( !(tmp & 0x4000) && tmp & 0x2000 )
              {	// only if 14 bits received
                rc5.data = tmp;
                rc5.received = TRUE;
              }
            tmp = 0;
          }

        if ( (rc5.bit ^ xRC5_IN) & (1<<xRC5) )
          {                                         // change detected
            rc5.bit = ~rc5.bit;				        // 0x00 -> 0xFF -> 0x00

            if( rc5.time < (PULSE_MIN / TICK_PERIOD) )			// to short
                tmp = 0;

            if( !tmp || (rc5.time > (PULSE_1_2 / TICK_PERIOD)) )
              {		                            // start or long pulse time
                if( !(tmp & 0x4000) )			    // not to many bits
                    tmp <<= 1;	                    // shift
                if( !(rc5.bit & (1<<xRC5)) )      // inverted bit
                    tmp |= 1;				// insert new bit
                rc5.time = 0;				// count next pulse time
              }
          }
        rc5.tmp = tmp;
      }

#elif (RC5_METHOD == 2)
//-----------------------------------------------------------------------------------
// Timinganalyse des RC5 Codes
// 1. run a filter to smooth the signal
// 2. Track Zeros and One, required eye opening 60%
//    we subtract EPS from both sides of halfbit

#define HALFBIT     (RC5TIME / 2 / TICK_PERIOD)     
#define EPS         (HALFBIT * 30L/ 100L)          // Tolerance corridor - eye opening

// security checks:
#if ((2*HALFBIT+EPS) > 255)
 #error: Overflow in RC5 receiver, please use a larger TICK_PERIOD
#endif
#if (HALFBIT < 7)
 #warning: resolution problems in RC5 receiver, please use a smaller TICK_PERIOD
 #warning: filter algorithm requires at least 7 samples per halfbit.
#endif
#if (EPS < 2)
 #warning: Error - no margin for transitions.
#endif

static inline void coding_error(void) __attribute__((always_inline));
void coding_error(void)
  {
    rc5.time = 0;
    rc5.state = IDLE;
  }

static inline void enter_bit(bool mybit) __attribute__((always_inline));
void enter_bit(bool mybit)
  {
     if (rc5.tmp & 0x2000) coding_error();      // we alread have 14 bits
     else
      {
        rc5.tmp = rc5.tmp << 1;
        if (mybit == 1) rc5.tmp++;
      }
  }

void rc5_sampler()
  {
    #if (USE_FILTER_ON_RC5 == TRUE)
        unsigned char temp;
        signed char filter_val = 0;

        temp = rc5.filter_data << 1;

        if (PIND & (1<<RC5_n))  temp |= 1;            // shift incoming data to filter

        sei();                                      // allow interrupts again
    
        rc5.filter_data = temp;

        if (temp & (1<<0)) filter_val++;              // lowpass over 7 samples
        if (temp & (1<<1)) filter_val++;
        if (temp & (1<<2)) filter_val++;
        if (temp & (1<<3)) filter_val++;
        if (temp & (1<<4)) filter_val++;
        if (temp & (1<<5)) filter_val++;
        if (temp & (1<<6)) filter_val++;

        filter_val -= 4;                               // filter_val now: [-4...3]
                                                       // sign bit is input state (already inverted)
        #define RC5_line (!(filter_val & (1<<7)))            // equal to level of port

    #else // (USE_FILTER_ON_RC5 == TRUE)

        #define RC5_line (PIND & (1<<RC5_n)) 

    #endif // (USE_FILTER_ON_RC5 == TRUE)


    #if (SHOW_IR_BITS_ON_LED == TRUE)
      { if (RC5_line) 
          {
            semaphor_set(C_LED);       // set flag for LED active
            LED_ON;
          }
        else
          {
            semaphor_get(C_LED);       // reset flag for LED active
            LED_OFF;
          }
      }
    #endif

    rc5.time++;                                    // time from last change

    switch (rc5.state)
      {
        case IDLE:
            if (RC5_line)                     
              {
                if (rc5.time == 0)              
                  {
                    rc5.time = 255;             // keep at max.
                  }
              }
            else
              {
                if (rc5.time == 0)              // do we come from pause
                  {
                    rc5.tmp = 0;
                    rc5.state = ONE_2ND_HALF;
                  }
                else
                  coding_error();
              }
            break;
        case ZERO_1ST_HALF:
            if (!RC5_line)
              {
                if (rc5.time > (HALFBIT+EPS))  coding_error();
              }
            else
              {
                if (rc5.time >= (HALFBIT-EPS))
                  {
                    rc5.state = ZERO_2ND_HALF;
                    rc5.time = 0;
                  }
                else coding_error();
              }
            break;
        case ZERO_2ND_HALF:
            if (RC5_line)
              {
                if (rc5.time > (2*HALFBIT+EPS)) 
                  {
                    enter_bit(0);
                    // we have coding error - either reached the end or real error
                    if( !(rc5.tmp & 0x4000) && (rc5.tmp & 0x2000) )
                      {	// only if 14 bits received
                        rc5.data = rc5.tmp;
                        rc5.received = TRUE;
                        rc5.time = 0;
                        rc5.state = IDLE;
                      }
                    else 
                      {
                        coding_error();
                      }
                  }
              }
            else
              {
                if (rc5.time >= (2*HALFBIT-EPS)) // must be a zero, followed by a one
                  {
                    rc5.state = ONE_2ND_HALF;
                    rc5.time = 0;
                    enter_bit(0);
                  }
                else if (rc5.time > (HALFBIT+EPS)) coding_error();
                else if (rc5.time >= (HALFBIT-EPS))  // must be a zero, followed by a zero
                  {
                    rc5.state = ZERO_1ST_HALF;
                    rc5.time = 0;
                    enter_bit(0);
                  }
                else coding_error();
              }
            break;
        
        case ONE_1ST_HALF:
            if (RC5_line)
              {
                if (rc5.time > (HALFBIT+EPS))
                  {
                    // we have coding error - either reached the end or real error
                    if( !(rc5.tmp & 0x4000) && (rc5.tmp & 0x2000) )
                      {	// only if 14 bits received
                        rc5.data = rc5.tmp;
                        rc5.received = TRUE;
                        rc5.time = 0;
                        rc5.state = IDLE;
                      }
                    else 
                      {
                        coding_error();
                      }
                  }
              }
            else
              {
                if (rc5.time >= (HALFBIT-EPS))
                  {
                    rc5.state = ONE_2ND_HALF;
                    rc5.time = 0;
                  }
                else coding_error();
              }
            break;
        case ONE_2ND_HALF:
            if (RC5_line)
              {
                if (rc5.time >= (2*HALFBIT-EPS))  // must be a one, followed by a zero
                  {
                    rc5.state = ZERO_2ND_HALF;
                    rc5.time = 0;
                    enter_bit(1);
                  }
                else if (rc5.time > (HALFBIT+EPS))  coding_error();
                  
                else if (rc5.time >= (HALFBIT-EPS))  // must be a one, followed by a one
                  {
                    rc5.state = ONE_1ST_HALF;
                    rc5.time = 0;
                    enter_bit(1);
                  }
                else coding_error();
              }
            else
              {
                if (rc5.time > (2*HALFBIT+EPS))
                  {
                    coding_error();
                  }
              }
            break;
      }
  }

#elif (RC5_METHOD == 3)

uint8_t impulse_prev = 0;          // This is the old pulse
uint8_t counter_impulse_1 = 0;
uint8_t flag_rx = 0;
uint8_t counter_impulse_0 = 0;
uint8_t impulse = 0;
uint64_t counter_ik = 0;
uint64_t code = 0;

void rc5_sampler(void)
  {
    uint8_t c;
    uint8_t str[35];
    uint64_t i;

    // Timer0 initialization

    TCNT0=0xFF;

    // Invert input signal

    if ((PIND & (1<<RC5_n))) impulse=0; else impulse=1;

    if ((impulse_prev==0)&&(impulse==1))
      {                                             //rising edge detection
        counter_impulse_1=0;                //reset counter

        if ((counter_impulse_0>3)&&(flag_rx==1)&&(counter_impulse_0<23))

          {  //if receiving symbols

            if(counter_impulse_0>12) // logical 1 received

            code |= (1<< 2);    //
            counter_ik++;

          };

      };

    if((impulse==1)&&(counter_impulse_1<50))

        //counting positive pulses
        counter_impulse_1++;

    if((impulse_prev==1)&&(impulse==0))
      {                                         //negative front

        counter_impulse_0=0;

        if(counter_impulse_1>30)
          {//one long synchro pulse

            code=0;
            counter_ik=0;
            flag_rx=1;
          }
      };

    if((impulse==0)&&(flag_rx==1)&(counter_impulse_1<150))

        //counting negative pulses
        counter_impulse_0++;

    if((counter_impulse_0>50)&&(flag_rx==1)&&(impulse==0))

      {//end of receive
        flag_rx=0;
        counter_impulse_0=0;
        counter_impulse_1=0;
        counter_ik=0;
        code=0;
      };

    impulse_prev=impulse;
  }

#else
  #warning Error: no RC5_METHOD selected
  void rc5_sampler() {};      // empty
#endif  // RC5 METHOD

#else   // RC5_ENABLED

void rc5_sampler() {};      // empty

#endif



//===============================================================================

// Long Bit fifo
//
// Put new bit to LSB of buffer[0], then shift complete buffer up one bit.
// Return top byte of Buffer.
// Note: declare buffer volatile, not to disturb the compiler optimization
//       The pointer to the buffer will be destroyed (I don't know how to tell it the
//       compiler), so multiple call with const pointers will dramatically fail :-(
//       (No problem, if used inside a ISR)
//
unsigned char bit_fifo(volatile uint8_t *buffer, uint8_t length, uint8_t newbit)
     __attribute__((always_inline));
unsigned char bit_fifo(volatile uint8_t *buffer, uint8_t length, uint8_t newbit)
  {
      uint8_t shifted;
	
      asm volatile (
         "ror %3"                          "\n"
         "BIT_FF_LOOP%=:"                  "\n"
         "ld %0,  %a1"                     "\n"
         "adc %0, %0"                      "\n"
         "st %a1+, %0"                     "\n"
         "dec %2"                          "\n"
         "brne BIT_FF_LOOP%="              "\n"
            : "=&r" (shifted)                               // outputs
            : "e"  (buffer), "r" (length), "r" (newbit)     // inputs
            : "memory"                                      // clobbered
      );
    return(shifted);
   }




//-----------------------------------------------------------------------------
// S88-Handler
//
// we use USI
// 
// a) 8-Bit mode (shift_size = 8)
//    USI could be used directly
//    states: 1. Shifting (no interaction)
//            2. Loading -> enable int on rising edge
//                          ISR will copy current_message to shifter
//
// b) 16-Bit mode (shift_size = 8..16)
//    USI is too short, needs a software extension
//    In this mode, we use USI only for two msb bit.
//    Interrupt on every clock: rising: 1. Shifting (sample data in)
//                                      2. Loading (copy current_message)
//                              falling: put out next Data
//                                       + prepare next bit (so copy 2 MSBs)
//                                       if there is only the MSB, Data Out
//                                       will have low immediately after clk low
//                                       (you can see timing then)
// S88 Reset is ignored
//-----------------------------------------------------------------------------


unsigned int shift_reg;
unsigned char shift_size = 8;   // may be 8 ... 16
// #define shift_size 16

// case b) Extension of shiftregister in SW
ISR(USI_OVERFLOW_vect)
  {
    unsigned char my_usi, temp;
    
    #if (SHOW_S88_ON_LED == TRUE)
      LED_ON;
    #endif

    USISR |= 0b00001111;      // set Counter to 15, to get an new interrupt

    #if (SHOW_S88_ON_LED == TRUE)
          if (!semaphor_query(C_LED)) LED_OFF;
    #endif



    if (S88_CLK)
      {
        // we got a rising edge:
        if (S88_LOAD)
          {
            shift_reg = current_message;
            semaphor_set(C_S88_Sent);           // tell main that data is sent

            my_usi = shift_reg >> (shift_size-8);
            // my_usi &= 0x80;  // if only MSB is written, you will see timing on a scope
            USIDR = my_usi;
          }
        else
          { 
            temp = USIDR & 0x01;                // read input
            shift_reg <<= 1;
            shift_reg += temp;
          }
      }
    else
      {
        // we got a falling edge, write next bit

         my_usi = shift_reg >> (shift_size-8);
         // my_usi = shift_reg >> (shift_size-2);  // then it could be used for any size from 2..16
         // my_usi = my_usi << 6;
         // my_usi &= 0x80;  // if only MSB is written, you will see timing on a scope   
         USIDR = my_usi;
      }
  }



#define KEY_PERIOD 10000L    // 10ms

#if ((KEY_PERIOD / TICK_PERIOD) > 120)
 #warning: overflow caused by a too large KEY_PERIOD
#endif


// volatile unsigned char aaa[]={0x01, 0xFF, 0x03, 0xAA, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
// volatile unsigned char *aptr;

int main(void)
  {
    signed char key_timer;

    init_main();

    if (JUMPER2_FITTED)
      {
        if (JUMPER1_FITTED)
          {
            keyboard_type = MATRIX_4x4;
            shift_size = 16;
          }
        else
          {
            keyboard_type = RC5_RECEIVER;
            shift_size = 16;                     // then rc5 receiver
          }
      }
    else
      {
        if (JUMPER1_FITTED)
          {
            keyboard_type = POLLIN_KEYBOARD;
            shift_size = 16;
          }
        else
          {
            keyboard_type = DIRECT;
            shift_size = 8;                     
          }
      }
 
    init_keyboard();

    sei();                                          // Global enable interrupts

    key_timer = timerval;

    while(1)
      {
        signed char temp;                   // trick down gcc: this forces char and smaller code
        temp = timerval - key_timer;
        switch (keyboard_type)
          {
            case RC5_RECEIVER:
                #if (RC5_ENABLED == TRUE)
                cli();
                current_message = rc5.data;
                sei();
                #endif
                #if (SHOW_IR_TOGGLE_ON_LED == TRUE)
                if (current_message & 0x0400)   // Toggle bit
                  {
                    semaphor_set(C_LED);       // set flag for LED active
                    LED_ON;
                  }
                else
                  {
                    semaphor_get(C_LED);       // reset flag for LED active
                    LED_OFF;
                  }
                #endif

                break;
            case MATRIX_4x4:
            case POLLIN_KEYBOARD:
            case DIRECT:
                if (temp > (KEY_PERIOD / TICK_PERIOD))
                  {
                    update_keys();
                    key_timer = timerval;
                    if (current_message)
                      {
                        semaphor_set(C_LED);       // set flag for LED active
                        LED_ON;
                      }
                    else
                      {
                        semaphor_get(C_LED);       // reset flag for LED active
                        LED_OFF;
                      }
                  } 
                break;
          }
      }
  }


//=========================================================================





//=========================================================================
//
// Simulationen:
//
// 0: keine Simulation
//
#if (SIMULATION == 1)

#endif // (SIMULATION == 1)
