//------------------------------------------------------------------------
//
// OpenDCC - S88-N - Keyboard and RC5 Receiver
//
// Copyright (c) 2007 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      twoport_keyboard.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2007-11-27 V0.1 start
//
//------------------------------------------------------------------------
//
// purpose:   keyboard and RC5 feedback system
//
// content:   
//             1. here keyboard routines for two different ports
//            
//
//------------------------------------------------------------------------

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <string.h>

#include "main.h"                // timerval, current_message

//------------------------------------------------------------------------------
// --------------- Port1:
#define KEY1_PORT           PINB         // This is the port where we read bits

#define KEY1_MASK           0b00001111   // the keyboard task runs only on
                                         // those ports defined in the mask
                                         // Ports not in the mask are ignored

#define KEY1_OFFSET         4            // shift keystroke result with this number
                                         // here: shift up by 4

#define KEY1_ACTIVE_TO_GND  TRUE         // TRUE: detect a keystroke if pin is grounded
                                         // FALSE: detect a keystroke if high is applied

unsigned char key1_state;                // static - save the previous state of keys.

unsigned char debounce1;                 // static - save a flag for "current in debounce"

signed char last_key_time1;              // last timerval

// --------------- Port1:
#define KEY2_PORT           PIND         // This is the port where we read bits

#define KEY2_MASK           0b01111000   // the keyboard task runs only on
                                         // those ports defined in the mask
                                         // Ports not in the mask are ignored

#define KEY2_OFFSET         -3           // shift keystroke result with this number

#define KEY2_ACTIVE_TO_GND  TRUE         // TRUE: detect a keystroke if pin is grounded
                                         // FALSE: detect a keystroke if high is applied

unsigned char key2_state;                // static - save the previous state of keys.

unsigned char debounce2;                 // static - save a flag for "current in debounce"

signed char last_key_time2;              // last timerval


//------------------------------------------------------------------------------
// code:

void init_keyboard_x(void)
  {
    debounce1 = 0;
    debounce2 = 0;
    #if (KEY1_ACTIVE_TO_GND == TRUE)
        key1_state = ~KEY1_PORT;  
    #else
        key1_state = KEY1_PORT;  
    #endif
    
    #if (KEY2_ACTIVE_TO_GND == TRUE)
        key2_state = ~KEY2_PORT;  
    #else
        key2_state = KEY2_PORT;  
    #endif

    last_key_time1 = timerval;                    // remember time 
    last_key_time2 = timerval;                    // remember time 
  }


//---------------------------------------------------------------------------
// keyboard(void)
// returns number corresponding to keystroke
//    0: Port (KEY_PORT), bit 0
//    1: Port (KEY_PORT), bit 1
//    ....
//    7: Port (KEY_PORT), bit 7
// 0xff: no keystroke detected
//
unsigned char keyboard1_poll(void)
  {
    unsigned char keys, xor;

    if ((char)(timerval - last_key_time1) < 1)  return(0xFF);

    last_key_time1++;

    #if (KEY1_ACTIVE_TO_GND == TRUE)
        keys = ~KEY1_PORT;  
    #else
        keys = KEY1_PORT;  
    #endif
    xor = keys ^ key1_state;
    xor = xor & KEY1_MASK;

    if (xor)
      {                         // something has changed
        unsigned char i, mask;

        mask = 1;
        for (i=0; i<8; i++)
          {
            if (xor & mask)
              {
                if (debounce1 & mask)
                  { // debounce over, now change state and report
                    debounce1 &= ~mask;
                    key1_state ^= mask;

                    if (keys & mask) return(i);
                  }
                else
                  { 
                    debounce1 |= mask;
                  }
              }
            mask = mask << 1;           
          }
      }
    else
      {
        debounce1 = 0;
      }
    return(0xFF);
  }


unsigned char keyboard2_poll(void)
  {
    unsigned char keys, xor;

    if ((char)(timerval - last_key_time2) < 1)  return(0xFF);

    last_key_time2++;

    #if (KEY2_ACTIVE_TO_GND == TRUE)
        keys = ~KEY2_PORT;  
    #else
        keys = KEY2_PORT;  
    #endif
    xor = keys ^ key2_state;
    xor = xor & KEY2_MASK;

    if (xor)
      {                         // something has changed
        unsigned char i, mask;

        mask = 1;
        for (i=0; i<8; i++)
          {
            if (xor & mask)
              {
                if (debounce2 & mask)
                  { // debounce over, now change state and report
                    debounce2 &= ~mask;
                    key2_state ^= mask;

                    if (keys & mask) return(i);
                  }
                else
                  { 
                    debounce2 |= mask;
                  }
              }
            mask = mask << 1;           
          }
      }
    else
      {
        debounce2 = 0;
      }
    return(0xFF);
  }



unsigned char keyboard(void)
  {
    unsigned char mykey;
    mykey = keyboard1_poll();
    if (mykey != 0xFF)
      {
         return((signed char) mykey + KEY1_OFFSET);
      }
    else
      {
        mykey = keyboard2_poll();
        if (mykey != 0xFF)
          {
             return((signed char) mykey + KEY2_OFFSET);
          }
      }
    return(0xFF);
  }










